SystemOrganization addCategory: #'Graphics-Text'!

Interval subclass: #TextLineInterval
	instanceVariableNames: 'internalSpaces paddingWidth lineHeight baseline'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!TextLineInterval commentStamp: '<historical>' prior: 0!
My instances specify the starting and stopping points in a String of a composed line. The step is always 1.!

!TextLineInterval class methodsFor: 'instance creation' stamp: 'lr 7/4/2009 10:42'!
start: startInteger stop: stopInteger internalSpaces: spacesInteger paddingWidth: padWidthInteger 
	"Answer an instance of me with the arguments as the start, stop points, 
	number of spaces in the line, and width of the padding."
	| newSelf |
	newSelf := super 
		from: startInteger
		to: stopInteger
		by: 1.
	^ newSelf 
		internalSpaces: spacesInteger
		paddingWidth: padWidthInteger! !

!TextLineInterval methodsFor: 'comparing' stamp: ''!
= line

	self species = line species
		ifTrue: [^((start = line first and: [stop = line last])
				and: [internalSpaces = line internalSpaces])
				and: [paddingWidth = line paddingWidth]]
		ifFalse: [^false]! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
baseline
	^ baseline! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
internalSpaces
	"Answer the number of spaces in the line."

	^internalSpaces! !

!TextLineInterval methodsFor: 'accessing' stamp: 'lr 7/4/2009 10:42'!
internalSpaces: spacesInteger 
	"Set the number of spaces in the line to be spacesInteger."
	internalSpaces := spacesInteger! !

!TextLineInterval methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
internalSpaces: spacesInteger paddingWidth: padWidthInteger 
	internalSpaces := spacesInteger.
	paddingWidth := padWidthInteger! !

!TextLineInterval methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
justifiedPadFor: spaceIndex 
	"Compute the width of pad for a given space in a line of justified text."
	| pad |
	internalSpaces = 0 ifTrue: [ ^ 0 ].
	pad := paddingWidth // internalSpaces.
	spaceIndex <= (paddingWidth \\ internalSpaces) 
		ifTrue: [ ^ pad + 1 ]
		ifFalse: [ ^ pad ]! !

!TextLineInterval methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
justifiedTabDeltaFor: spaceIndex 
	"Compute the delta for a tab in a line of justified text, so tab falls 
	somewhere plausible when line is justified."
	| pad extraPad |
	internalSpaces = 0 ifTrue: [ ^ 0 ].
	pad := paddingWidth // internalSpaces.
	extraPad := paddingWidth \\ internalSpaces.
	spaceIndex <= extraPad 
		ifTrue: [ ^ spaceIndex * (pad + 1) ]
		ifFalse: [ ^ extraPad * (pad + 1) + ((spaceIndex - extraPad) * pad) ]! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
lineHeight
	^ lineHeight! !

!TextLineInterval methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
lineHeight: height baseline: ascent 
	lineHeight := height.
	baseline := ascent! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
paddingWidth
	"Answer the amount of space to be added to the font."

	^paddingWidth! !

!TextLineInterval methodsFor: 'accessing' stamp: 'lr 7/4/2009 10:42'!
paddingWidth: padWidthInteger 
	"Set the amount of space to be added to the font to be padWidthInteger."
	paddingWidth := padWidthInteger! !

!TextLineInterval methodsFor: 'updating' stamp: 'lr 7/4/2009 10:42'!
slide: delta 
	"Change the starting and stopping points of the line by delta."
	start := start + delta.
	stop := stop + delta! !

!TextLineInterval methodsFor: 'accessing' stamp: 'lr 7/4/2009 10:42'!
stop: stopInteger 
	"Set the stopping point in the string of the line to be stopInteger."
	stop := stopInteger! !

Rectangle subclass: #CharacterBlock
	instanceVariableNames: 'stringIndex text textLine'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CharacterBlock commentStamp: 'StephaneDucasse 5/18/2010 16:00' prior: 0!
I describe the location of one character displayed on the screen. My instances are used to return the results of methods:
	Paragraph characterBlockAtPoint: aPoint and
	Paragraph characterBlockForIndex: stringIndex.
Any recomposition or movement of a Paragraph can make the information I store stale.

text (Text): The text where my character is from
stringIndex (Integer): The index of my character in the text, starting from 1
textLine (TextLine): The displayed line my character is on
origin (Point): The top-left corner of the area allocated for displaying my
		character's glyph, in pixels, counting right then down from the
		top-left corner of the text display area, and starting from 0@0
corner (Point): The bottom-right corner of the area allocated for displaying my
		character's glyph, in pixels, counting right then down from the
		top-left corner of the text display area, and starting from 0@0
!

!CharacterBlock methodsFor: 'comparing' stamp: ''!
< aCharacterBlock 
	"Answer whether the string index of the receiver precedes that of 
	aCharacterBlock."

	^stringIndex < aCharacterBlock stringIndex! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
<= aCharacterBlock 
	"Answer whether the string index of the receiver does not come after that 
	of aCharacterBlock."

	^(self > aCharacterBlock) not! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
= aCharacterBlock

	self species = aCharacterBlock species
		ifTrue: [^stringIndex = aCharacterBlock stringIndex]
		ifFalse: [^false]! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
> aCharacterBlock 
	"Answer whether the string index of the receiver comes after that of 
	aCharacterBlock."

	^aCharacterBlock < self! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
>= aCharacterBlock 
	"Answer whether the string index of the receiver does not precede that of 
	aCharacterBlock."

	^(self < aCharacterBlock) not! !

!CharacterBlock methodsFor: 'comparing' stamp: 'th 9/17/2002 11:54'!
max: aCharacterBlock
	aCharacterBlock ifNil:[^self].
	^aCharacterBlock > self
		ifTrue:[ aCharacterBlock]
		ifFalse:[self].! !

!CharacterBlock methodsFor: 'comparing' stamp: 'th 9/17/2002 11:54'!
min: aCharacterBlock
	aCharacterBlock ifNil:[^self].
	^aCharacterBlock < self
		ifTrue:[ aCharacterBlock]
		ifFalse:[self].! !

!CharacterBlock methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
moveBy: aPoint 
	"Change the corner positions of the receiver so that its area translates by 
	the amount defined by the argument, aPoint."
	origin := origin + aPoint.
	corner := corner + aPoint! !

!CharacterBlock methodsFor: 'printing' stamp: 'di 12/2/97 19:15'!
printOn: aStream

	aStream nextPutAll: 'a CharacterBlock with index '.
	stringIndex printOn: aStream.
	(text ~~ nil and: [text size> 0 and: [stringIndex between: 1 and: text size]])
		ifTrue: [aStream nextPutAll: ' and character '.
				(text at: stringIndex) printOn: aStream].
	aStream nextPutAll: ' and rectangle '.
	super printOn: aStream.
	textLine ifNotNil: [aStream cr; nextPutAll: ' in '.
				textLine printOn: aStream].
! !

!CharacterBlock methodsFor: 'accessing' stamp: ''!
stringIndex
	"Answer the position of the receiver in the string it indexes."

	^stringIndex! !

!CharacterBlock methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
stringIndex: anInteger text: aText topLeft: topLeft extent: extent 
	stringIndex := anInteger.
	text := aText.
	super 
		setOrigin: topLeft
		corner: topLeft + extent! !

!CharacterBlock methodsFor: 'accessing' stamp: 'di 12/2/97 14:33'!
textLine
	^ textLine! !

!CharacterBlock methodsFor: 'accessing' stamp: 'lr 7/4/2009 10:42'!
textLine: aLine 
	textLine := aLine! !

Object subclass: #CharacterScanner
	instanceVariableNames: 'destX lastIndex xTable destY stopConditions text textStyle alignment leftMargin rightMargin font line runStopIndex spaceCount spaceWidth emphasisCode kern indentationLevel wantsColumnBreaks pendingKernX'
	classVariableNames: 'DefaultStopConditions NilCondition PaddedSpaceCondition SpaceCondition'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CharacterScanner commentStamp: '<historical>' prior: 0!
My instances hold the state associated with scanning text. My subclasses scan characters for specified purposes, such as computing a CharacterBlock or placing characters into Forms.!

CharacterScanner subclass: #CharacterBlockScanner
	instanceVariableNames: 'characterPoint characterIndex lastCharacter lastCharacterExtent lastSpaceOrTabExtent nextLeftMargin specialWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!CharacterBlockScanner commentStamp: '<historical>' prior: 0!
My instances are used to scan text to compute the CharacterBlock for a character specified by its index in the text or its proximity to the cursor location.!

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
buildCharacterBlockIn: para 
	"handle nullText"
	| lineIndex runLength lineStop done stopCondition |
	(para numberOfLines = 0 or: [ text size = 0 ]) ifTrue: 
		[ ^ CharacterBlock new 
			stringIndex: 1
			text: para text
			topLeft: (para 
				leftMarginForDisplayForLine: 1
				alignment: (alignment ifNil: 
					[ "like being off end of string"
					textStyle alignment ])) @ para compositionRectangle top
			extent: 0 @ textStyle lineGrid ].
	"find the line"
	lineIndex := para lineIndexOfTop: characterPoint y.
	destY := para topAtLineIndex: lineIndex.
	line := para lines at: lineIndex.
	lastIndex := line first.
	rightMargin := para rightMarginForDisplay.
	self setStopConditions.	" also loads the font and loads all emphasis attributes "
	(lineIndex = para numberOfLines and: [ destY + line lineHeight < characterPoint y ]) 
		ifTrue: 
			[ "if beyond lastLine, force search to last character"
			self characterPointSetX: rightMargin ]
		ifFalse: 
			[ characterPoint y < para compositionRectangle top ifTrue: 
				[ "force search to first line"
				characterPoint := para compositionRectangle topLeft ].
			characterPoint x > rightMargin ifTrue: [ self characterPointSetX: rightMargin ] ].
	destX := leftMargin := para 
		leftMarginForDisplayForLine: lineIndex
		alignment: (alignment ifNil: [ textStyle alignment ]).
	nextLeftMargin := para 
		leftMarginForDisplayForLine: lineIndex + 1
		alignment: (alignment ifNil: [ textStyle alignment ]).
	lastIndex := line first.
	self setStopConditions.	"also sets font"
	runLength := text runLengthFor: line first.
	characterIndex == nil 
		ifTrue: [ lineStop := line last	"characterBlockAtPoint" ]
		ifFalse: [ lineStop := characterIndex	"characterBlockForIndex" ].
	(runStopIndex := lastIndex + (runLength - 1)) > lineStop ifTrue: [ runStopIndex := lineStop ].
	lastCharacterExtent := 0 @ line lineHeight.
	spaceCount := 0.
	done := false.
	self handleIndentation.
	[ done ] whileFalse: 
		[ stopCondition := self 
			scanCharactersFrom: lastIndex
			to: runStopIndex
			in: text string
			rightX: characterPoint x
			stopConditions: stopConditions
			kern: kern.

		"see setStopConditions for stopping conditions for character block 	operations."
		self lastCharacterExtentSetX: (font widthOf: (text at: lastIndex)).
		(self perform: stopCondition) ifTrue: 
			[ characterIndex == nil 
				ifTrue: 
					[ "characterBlockAtPoint"
					^ CharacterBlock new 
						stringIndex: lastIndex
						text: text
						topLeft: characterPoint + (font descentKern @ 0)
						extent: lastCharacterExtent ]
				ifFalse: 
					[ "characterBlockForIndex"
					^ CharacterBlock new 
						stringIndex: lastIndex
						text: text
						topLeft: characterPoint + ((font descentKern - kern) @ 0)
						extent: lastCharacterExtent ] ] ]! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
characterBlockAtPoint: aPoint in: aParagraph 
	"Answer a CharacterBlock for character in aParagraph at point aPoint. It 
	is assumed that aPoint has been transformed into coordinates appropriate 
	to the text's destination form rectangle and the composition rectangle."
	self 
		initializeFromParagraph: aParagraph
		clippedBy: aParagraph clippingRectangle.
	characterPoint := aPoint.
	^ self buildCharacterBlockIn: aParagraph! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'nk 11/22/2004 14:32'!
characterBlockAtPoint: aPoint index: index in: textLine
	"This method is the Morphic characterBlock finder.  It combines
	MVC's characterBlockAtPoint:, -ForIndex:, and buildCharcterBlock:in:"
	| runLength lineStop done stopCondition |
	line := textLine.
	rightMargin := line rightMargin.
	lastIndex := line first.
	self setStopConditions.		"also sets font"
	characterIndex := index.  " == nil means scanning for point"
	characterPoint := aPoint.
	(characterPoint isNil or: [characterPoint y > line bottom])
		ifTrue: [characterPoint := line bottomRight].
	(text isEmpty or: [(characterPoint y < line top or: [characterPoint x < line left])
				or: [characterIndex notNil and: [characterIndex < line first]]])
		ifTrue:	[^ (CharacterBlock new stringIndex: line first text: text
					topLeft: line leftMargin@line top extent: 0 @ textStyle lineGrid)
					textLine: line].
	destX := leftMargin := line leftMarginForAlignment: alignment.
	destY := line top.
	runLength := text runLengthFor: line first.
	characterIndex
		ifNotNil:	[lineStop := characterIndex  "scanning for index"]
		ifNil:	[lineStop := line last  "scanning for point"].
	runStopIndex := lastIndex + (runLength - 1) min: lineStop.
	lastCharacterExtent := 0 @ line lineHeight.
	spaceCount := 0.

	done  := false.
	[done] whileFalse:
		[stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
			in: text string rightX: characterPoint x
			stopConditions: stopConditions kern: kern.
		"see setStopConditions for stopping conditions for character block 	operations."
		self lastCharacterExtentSetX: (specialWidth
			ifNil: [font widthOf: (text at: lastIndex)]
			ifNotNil: [specialWidth]).
		(self perform: stopCondition) ifTrue:
			[characterIndex
				ifNil: [
					"Result for characterBlockAtPoint: "
					(stopCondition ~~ #cr and: [ lastIndex == line last
						and: [ aPoint x > ((characterPoint x) + (lastCharacterExtent x / 2)) ]])
							ifTrue: [ "Correct for right half of last character in line"
								^ (CharacterBlock new stringIndex: lastIndex + 1
										text: text
										topLeft: characterPoint + (lastCharacterExtent x @ 0) + (font descentKern @ 0)
										extent:  0 @ lastCharacterExtent y)
									textLine: line ].
						^ (CharacterBlock new stringIndex: lastIndex
							text: text topLeft: characterPoint + (font descentKern @ 0)
							extent: lastCharacterExtent - (font baseKern @ 0))
									textLine: line]
				ifNotNil: ["Result for characterBlockForIndex: "
						^ (CharacterBlock new stringIndex: characterIndex
							text: text topLeft: characterPoint + ((font descentKern) - kern @ 0)
							extent: lastCharacterExtent)
									textLine: line]]]! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
characterBlockForIndex: targetIndex in: aParagraph 
	"Answer a CharacterBlock for character in aParagraph at targetIndex. The 
	coordinates in the CharacterBlock will be appropriate to the intersection 
	of the destination form rectangle and the composition rectangle."
	self 
		initializeFromParagraph: aParagraph
		clippedBy: aParagraph clippingRectangle.
	characterIndex := targetIndex.
	characterPoint := aParagraph rightMarginForDisplay @ (aParagraph topAtLineIndex: (aParagraph lineIndexOfCharacterIndex: characterIndex)).
	^ self buildCharacterBlockIn: aParagraph! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
characterPointSetX: xVal 
	characterPoint := xVal @ characterPoint y! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'nice 11/30/2009 23:04'!
cr
	"Answer a CharacterBlock that specifies the current location of the mouse 
	relative to a carriage return stop condition that has just been 
	encountered. The ParagraphEditor convention is to denote selections by 
	CharacterBlocks, sometimes including the carriage return (cursor is at 
	the end) and sometimes not (cursor is in the middle of the text)."
	((characterIndex ~= nil and: [ characterIndex > text size ]) or: 
		[ line last = text size and: [ destY + line lineHeight < characterPoint y ] ]) ifTrue: 
		[ "When off end of string, give data for next character"
		destY := destY + line lineHeight.
		lastCharacter := nil.
		characterPoint := (nextLeftMargin ifNil: [ leftMargin ]) @ destY.
		(lastIndex < text size and: [(text at: lastIndex) = CR and: [(text at: lastIndex+1) = Character lf]])
			ifTrue: [lastIndex := lastIndex + 2]
			ifFalse: [lastIndex := lastIndex + 1].
		self lastCharacterExtentSetX: 0.
		^ true ].
	lastCharacter := CR.
	characterPoint := destX @ destY.
	self lastCharacterExtentSetX: rightMargin - destX.
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 12:49'!
crossedX
	"Text display has wrapping. The scanner just found a character past the x 
	location of the cursor. We know that the cursor is pointing at a character 
	or before one."

	| leadingTab currentX |
	characterIndex == nil ifFalse: [
		"If the last character of the last line is a space,
		and it crosses the right margin, then locating
		the character block after it is impossible without this hack."
		characterIndex > text size ifTrue: [
			lastIndex := characterIndex.
			characterPoint := (nextLeftMargin ifNil: [leftMargin]) @ (destY + line lineHeight).
			^true]].
	characterPoint x <= (destX + (lastCharacterExtent x // 2))
		ifTrue:	[lastCharacter := (text at: lastIndex).
				characterPoint := destX @ destY.
				^true].
	lastIndex >= line last 
		ifTrue:	[lastCharacter := (text at: line last).
				characterPoint := destX @ destY.
				^true].

	"Pointing past middle of a character, return the next character."
	lastIndex := lastIndex + 1.
	lastCharacter := text at: lastIndex.
	currentX := destX + lastCharacterExtent x + kern.
	self lastCharacterExtentSetX: (font widthOf: lastCharacter).
	characterPoint := currentX @ destY.
	lastCharacter = Space ifFalse: [^ true].

	"Yukky if next character is space or tab."
	alignment = Justified ifTrue:
		[self lastCharacterExtentSetX:
			(lastCharacterExtent x + 	(line justifiedPadFor: (spaceCount + 1) font: font)).
		^ true].

	true ifTrue: [^ true].
	"NOTE:  I find no value to the following code, and so have defeated it - DI"

	"See tabForDisplay for illumination on the following awfulness."
	leadingTab := true.
	line first to: lastIndex - 1 do:
		[:index | (text at: index) ~= Tab ifTrue: [leadingTab := false]].
	(alignment ~= Justified or: [leadingTab])
		ifTrue:	[self lastCharacterExtentSetX: (textStyle nextTabXFrom: currentX
					leftMargin: leftMargin rightMargin: rightMargin) -
						currentX]
		ifFalse:	[self lastCharacterExtentSetX:  (((currentX + (textStyle tabWidth -
						(line justifiedTabDeltaFor: spaceCount))) -
							currentX) max: 0)].
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
endOfRun
	"Before arriving at the cursor location, the selection has encountered an 
	end of run. Answer false if the selection continues, true otherwise. Set 
	up indexes for building the appropriate CharacterBlock."
	| runLength lineStop |
	(((characterIndex ~~ nil and: [ runStopIndex < characterIndex and: [ runStopIndex < text size ] ]) or: [ characterIndex == nil and: [ lastIndex < line last ] ]) or: 
		[ lastIndex < line last and: 
			[ (text at: lastIndex) leadingChar ~= (text at: lastIndex + 1) leadingChar and: [ lastIndex ~= characterIndex ] ] ]) ifTrue: 
		[ "We're really at the end of a real run."
		runLength := text runLengthFor: (lastIndex := lastIndex + 1).
		characterIndex ~~ nil 
			ifTrue: [ lineStop := characterIndex	"scanning for index" ]
			ifFalse: [ lineStop := line last	"scanning for point" ].
		(runStopIndex := lastIndex + (runLength - 1)) > lineStop ifTrue: [ runStopIndex := lineStop ].
		self setStopConditions.
		^ false ].
	lastCharacter := text at: lastIndex.
	characterPoint := destX @ destY.
	((lastCharacter = Space and: [ alignment = Justified ]) or: [ lastCharacter = Tab and: [ lastSpaceOrTabExtent notNil ] ]) ifTrue: [ lastCharacterExtent := lastSpaceOrTabExtent ].
	characterIndex ~~ nil ifTrue: 
		[ "If scanning for an index and we've stopped on that index,
				then we back destX off by the width of the character stopped on
				(it will be pointing at the right side of the character) and return"
		runStopIndex = characterIndex ifTrue: 
			[ self characterPointSetX: destX - lastCharacterExtent x.
			^ true ].
		"Otherwise the requested index was greater than the length of the
				string.  Return string size + 1 as index, indicate further that off the
				string by setting character to nil and the extent to 0."
		lastIndex := lastIndex + 1.
		lastCharacter := nil.
		self lastCharacterExtentSetX: 0.
		^ true ].

	"Scanning for a point and either off the end of the line or off the end of the string."
	runStopIndex = text size ifTrue: 
		[ "off end of string"
		lastIndex := lastIndex + 1.
		lastCharacter := nil.
		self lastCharacterExtentSetX: 0.
		^ true ].
	"just off end of line without crossing x"
	lastIndex := lastIndex + 1.
	^ true! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
indentationLevel: anInteger 
	super indentationLevel: anInteger.
	nextLeftMargin := leftMargin.
	indentationLevel timesRepeat: 
		[ nextLeftMargin := textStyle 
			nextTabXFrom: nextLeftMargin
			leftMargin: leftMargin
			rightMargin: rightMargin ]! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
lastCharacterExtentSetX: xVal 
	lastCharacterExtent := xVal @ lastCharacterExtent y! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
lastSpaceOrTabExtentSetX: xVal 
	lastSpaceOrTabExtent := xVal @ lastSpaceOrTabExtent y! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 12:50'!
paddedSpace
	"When the line is justified, the spaces will not be the same as the font's 
	space character. A padding of extra space must be considered in trying 
	to find which character the cursor is pointing at. Answer whether the 
	scanning has crossed the cursor."

	| pad |
	pad := 0.
	spaceCount := spaceCount + 1.
	pad := line justifiedPadFor: spaceCount font: font.
	lastSpaceOrTabExtent := lastCharacterExtent copy.
	self lastSpaceOrTabExtentSetX:  spaceWidth + pad.
	(destX + lastSpaceOrTabExtent x)  >= characterPoint x
		ifTrue: [lastCharacterExtent := lastSpaceOrTabExtent copy.
				^self crossedX].
	lastIndex := lastIndex + 1.
	destX := destX + lastSpaceOrTabExtent x.
	^ false
! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	"Workaround: The following should really use #textAnchorType"
	anchoredMorph relativeTextAnchorPosition ifNotNil: [ ^ true ].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: [ ^ false ].
	specialWidth := anchoredMorph width.
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
setFont
	specialWidth := nil.
	super setFont! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 10/18/2004 14:30'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont.
	self setConditionArray: (alignment = Justified ifTrue: [#paddedSpace]).
! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
tab
	| currentX |
	currentX := (alignment == Justified and: [ self leadingTab not ]) 
		ifTrue: 
			[ "imbedded tabs in justified text are weird"
			destX + (textStyle tabWidth - (line justifiedTabDeltaFor: spaceCount)) max: destX ]
		ifFalse: 
			[ textStyle 
				nextTabXFrom: destX
				leftMargin: leftMargin
				rightMargin: rightMargin ].
	lastSpaceOrTabExtent := lastCharacterExtent copy.
	self lastSpaceOrTabExtentSetX: (currentX - destX max: 0).
	currentX >= characterPoint x ifTrue: 
		[ lastCharacterExtent := lastSpaceOrTabExtent copy.
		^ self crossedX ].
	destX := currentX.
	lastIndex := lastIndex + 1.
	^ false! !

!CharacterScanner class methodsFor: 'class initialization' stamp: 'nice 3/8/2010 11:51'!
initialize
"
	CharacterScanner initialize
"
	| a |
	a := TextStopConditions new.
	a at: 1 + 1 put: #embeddedObject.
	a at: Tab asciiValue + 1 put: #tab.
	a at: CR asciiValue + 1 put: #cr.
	a at: Character lf asciiValue + 1 put: #cr.
	
	NilCondition := a copy.
	DefaultStopConditions := a copy.

	PaddedSpaceCondition := a copy.
	PaddedSpaceCondition at: Space asciiValue + 1 put: #paddedSpace.
	
	SpaceCondition := a copy.
	SpaceCondition at: Space asciiValue + 1 put: #space.
! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
addEmphasis: code 
	"Set the bold-ital-under-strike emphasis."
	emphasisCode := emphasisCode bitOr: code! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
addKern: kernDelta 
	"Set the current kern amount."
	kern := kern + kernDelta! !

!CharacterScanner methodsFor: 'scanning' stamp: 'nice 2/28/2010 18:17'!
basicScanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta
	"Primitive. This is the inner loop of text display--but see 
	scanCharactersFrom: to:rightX: which would get the string, 
	stopConditions and displaying from the instance. March through source 
	String from startIndex to stopIndex. If any character is flagged with a 
	non-nil entry in stops, then return the corresponding value. Determine 
	width of each character from xTable, indexed by map. 
	If dextX would exceed rightX, then return stops at: 258. 
	Advance destX by the width of the character. If stopIndex has been
	reached, then return stops at: 257. Optional. 
	See Object documentation whatIsAPrimitive."
	| ascii nextDestX char  floatDestX widthAndKernedWidth nextChar atEndOfRun |
	<primitive: 103>
	lastIndex := startIndex.
	floatDestX := destX.
	widthAndKernedWidth := Array new: 2.
	atEndOfRun := false.
	[lastIndex <= stopIndex]
		whileTrue: 
			[char := (sourceString at: lastIndex).
			ascii := char asciiValue + 1.
			(stops at: ascii) == nil ifFalse: [^stops at: ascii].
			"Note: The following is querying the font about the width
			since the primitive may have failed due to a non-trivial
			mapping of characters to glyphs or a non-existing xTable."
			nextChar := (lastIndex + 1 <= stopIndex) 
				ifTrue:[sourceString at: lastIndex + 1]
				ifFalse:[
					atEndOfRun := true.
					"if there is a next char in sourceString, then get the kern 
					and store it in pendingKernX"
					lastIndex + 1 <= sourceString size
						ifTrue:[sourceString at: lastIndex + 1]
						ifFalse:[	nil]].
			font 
				widthAndKernedWidthOfLeft: char 
				right: nextChar
				into: widthAndKernedWidth.
			nextDestX := floatDestX + (widthAndKernedWidth at: 1).
			nextDestX > rightX ifTrue: [^stops crossedX].
			floatDestX := floatDestX + kernDelta + (widthAndKernedWidth at: 2).
			atEndOfRun 
				ifTrue:[
					pendingKernX := (widthAndKernedWidth at: 2) - (widthAndKernedWidth at: 1).
					floatDestX := floatDestX - pendingKernX].
			destX := floatDestX.
			lastIndex := lastIndex + 1].
	lastIndex := stopIndex.
	^stops endOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'tween 4/6/2007 11:16'!
columnBreak

	pendingKernX := 0.
	^true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
embeddedObject
	| savedIndex |
	savedIndex := lastIndex.
	text 
		attributesAt: lastIndex
		do: 
			[ :attr | 
			attr anchoredMorph ifNotNil: 
				[ "Following may look strange but logic gets reversed.
			If the morph fits on this line we're not done (return false for true) 
			and if the morph won't fit we're done (return true for false)"
				(self placeEmbeddedObject: attr anchoredMorph) ifFalse: [ ^ true ] ] ].
	lastIndex := savedIndex + 1.	"for multiple(!!) embedded morphs"
	^ false! !

!CharacterScanner methodsFor: 'scanning' stamp: 'hmm 7/15/2000 22:40'!
handleIndentation
	self indentationLevel timesRepeat: [
		self plainTab]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 5/17/2000 18:20'!
indentationLevel
	"return the number of tabs that are currently being placed at the beginning of each line"
	^indentationLevel ifNil:[0]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
indentationLevel: anInteger 
	"set the number of tabs to put at the beginning of each line"
	indentationLevel := anInteger! !

!CharacterScanner methodsFor: 'initialize' stamp: 'alain.plantec 5/28/2009 09:44'!
initialize
	super initialize.
	destX := destY := leftMargin := 0.! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
initializeFromParagraph: aParagraph clippedBy: clippingRectangle 
	text := aParagraph text.
	textStyle := aParagraph textStyle! !

!CharacterScanner methodsFor: 'initialize' stamp: 'ul 3/8/2010 04:44'!
initializeStringMeasurer

	stopConditions := TextStopConditions new
! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'yo 12/18/2002 12:32'!
isBreakableAtIndex: index

	^ (EncodedCharSet at: ((text at: index) leadingChar + 1)) isBreakableAt: index in: text.
! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 1/8/2000 14:23'!
leadingTab
	"return true if only tabs lie to the left"
	line first to: lastIndex do:
		[:i | (text at: i) == Tab ifFalse: [^ false]].
	^ true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
measureString: aString inFont: aFont from: startIndex to: stopIndex 
	"WARNING: In order to use this method the receiver has to be set up using #initializeStringMeasurer"
	destX := destY := lastIndex := 0.
	xTable := aFont xTable.
	self 
		scanCharactersFrom: startIndex
		to: stopIndex
		in: aString
		rightX: 999999
		stopConditions: stopConditions
		kern: 0.
	^ destX! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	"Place the anchoredMorph or return false if it cannot be placed.
	In any event, advance destX by its width."
	"Workaround: The following should really use #textAnchorType"
	| w |
	anchoredMorph relativeTextAnchorPosition ifNotNil: [ ^ true ].
	destX := destX + (w := anchoredMorph width).
	(destX > rightMargin and: [ leftMargin + w <= rightMargin ]) ifTrue: 
		[ "Won't fit, but would on next line"
		^ false ].
	lastIndex := lastIndex + 1.
	self setFont.	"Force recalculation of emphasis for next run"
	^ true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'nice 3/16/2010 20:38'!
plainTab
	"This is the basic method of adjusting destX for a tab."
	destX := (alignment = Justified and: [ self leadingTab not ]) 
		ifTrue: 
			[ "embedded tabs in justified text are weird"
			destX + (textStyle tabWidth - (line justifiedTabDeltaFor: spaceCount)) max: destX ]
		ifFalse: 
			[ textStyle 
				nextTabXFrom: destX
				leftMargin: leftMargin
				rightMargin: rightMargin ]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'nice 3/6/2010 14:47'!
scanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta 
	| startEncoding selector |
	sourceString isByteString ifTrue: 
		[ ^ self 
			basicScanCharactersFrom: startIndex
			to: stopIndex
			in: sourceString
			rightX: rightX
			stopConditions: stops
			kern: kernDelta ].
	sourceString isWideString ifTrue: 
		[ startIndex > stopIndex ifTrue: 
			[ lastIndex := stopIndex.
			^ stops endOfRun ].
		startEncoding := (sourceString at: startIndex) leadingChar.
		selector := (EncodedCharSet charsetAt: startEncoding) scanSelector.
		^ self 
			perform: selector
			withArguments: (Array 
					with: startIndex
					with: stopIndex
					with: sourceString
					with: rightX
					with: stops
					with: kernDelta) ].
	^ stops endOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'StephaneDucasse 11/26/2010 19:31'!
scanJapaneseCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta 
	| ascii encoding f nextDestX maxAscii startEncoding |
	lastIndex := startIndex.
	lastIndex > stopIndex ifTrue: 
		[ lastIndex := stopIndex.
		^ stops endOfRun ].
	startEncoding := (sourceString at: startIndex) leadingChar.
	font ifNil: [ font := (TextSharedInformation at: #DefaultMultiStyle) fontArray at: 1 ].
	((font isMemberOf: StrikeFontSet) or: [ font isKindOf: TTCFontSet ]) 
		ifTrue: 
			[ maxAscii := font maxAsciiFor: startEncoding.
			f := font fontArray at: startEncoding + 1.
			"xTable := f xTable.
		maxAscii := xTable size - 2."
			spaceWidth := f widthOf: Space ]
		ifFalse: [ maxAscii := font maxAscii ].
	[ lastIndex <= stopIndex ] whileTrue: 
		[ encoding := (sourceString at: lastIndex) leadingChar.
		encoding ~= startEncoding ifTrue: 
			[ lastIndex := lastIndex - 1.
			^ stops endOfRun ].
		ascii := (sourceString at: lastIndex) charCode.
		ascii > maxAscii ifTrue: [ ascii := maxAscii ].
		(encoding = 0 and: [ascii < stops size and: [ (stopConditions at: ascii + 1) ~~ nil ]]) ifTrue: [ ^ stops at: ascii + 1 ].
		nextDestX := destX + (font widthOf: (sourceString at: lastIndex)).
		nextDestX > rightX ifTrue: [ ^ stops crossedX ].
		destX := nextDestX + kernDelta.
		"destX printString displayAt: 0@(lastIndex*20)."
		lastIndex := lastIndex + 1 ].
	lastIndex := stopIndex.
	^ stops endOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'nice 2/28/2010 19:10'!
scanMultiCharactersCombiningFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta
	"Implement a workaround for unicode composing.
	a MultiCharacterScanner should better be used to handle combination."

	^self scanMultiCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'StephaneDucasse 11/26/2010 19:31'!
scanMultiCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta 
	| ascii encoding f nextDestX maxAscii startEncoding |
	lastIndex := startIndex.
	lastIndex > stopIndex ifTrue: 
		[ lastIndex := stopIndex.
		^ stops endOfRun ].
	startEncoding := (sourceString at: startIndex) leadingChar.
	font ifNil: [ font := (TextSharedInformation at: #DefaultMultiStyle) fontArray at: 1 ].
	((font isMemberOf: StrikeFontSet) or: [ font isKindOf: TTCFontSet ]) 
		ifTrue: 
			[ maxAscii := font maxAsciiFor: startEncoding.
			f := font fontArray at: startEncoding + 1.
			spaceWidth := f widthOf: Space ]
		ifFalse: [ maxAscii := font maxAscii ].
	[ lastIndex <= stopIndex ] whileTrue: 
		[ encoding := (sourceString at: lastIndex) leadingChar.
		encoding ~= startEncoding ifTrue: 
			[ lastIndex := lastIndex - 1.
			^ stops endOfRun ].
		ascii := (sourceString at: lastIndex) charCode.
		ascii > maxAscii ifTrue: [ ascii := maxAscii ].
		(encoding = 0 and: [ascii < stops size and: [ (stopConditions at: ascii + 1) ~~ nil ]]) ifTrue: [ ^ stops at: ascii + 1 ].
		nextDestX := destX + (font widthOf: (sourceString at: lastIndex)).
		nextDestX > rightX ifTrue: [ ^ stops crossedX ].
		destX := nextDestX + kernDelta.
		"destX printString displayAt: 0@(lastIndex*20)."
		lastIndex := lastIndex + 1 ].
	lastIndex := stopIndex.
	^ stops endOfRun! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setActualFont: aFont 
	"Set the basal font to an isolated font reference."
	font := aFont! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setAlignment: style 
	alignment := style! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setConditionArray: aSymbol 
	aSymbol == #paddedSpace ifTrue: [ ^ stopConditions := PaddedSpaceCondition copy ].
	aSymbol == #space ifTrue: [ ^ stopConditions := SpaceCondition copy ].
	aSymbol == nil ifTrue: [ ^ stopConditions := NilCondition copy ].
	self error: 'undefined stopcondition for space character'! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'tween 4/6/2007 09:59'!
setFont
	| priorFont |
	"Set the font and other emphasis."
	priorFont := font.
	text == nil ifFalse:[
		emphasisCode := 0.
		kern := 0.
		indentationLevel := 0.
		alignment := textStyle alignment.
		font := nil.
		(text attributesAt: lastIndex forStyle: textStyle)
			do: [:att | att emphasizeScanner: self]].
	font == nil ifTrue:
		[self setFont: textStyle defaultFontIndex].
	font := font emphasized: emphasisCode.
	priorFont 
		ifNotNil: [
			font = priorFont 
				ifTrue:[
					"font is the same, perhaps the color has changed?
					We still want kerning between chars of the same
					font, but of different color. So add any pending kern to destX"
					destX := destX + (pendingKernX ifNil:[0])].
			destX := destX + priorFont descentKern].
	pendingKernX := 0. "clear any pending kern so there is no danger of it being added twice"
	destX := destX - font descentKern.
	"NOTE: next statement should be removed when clipping works"
	leftMargin ifNotNil: [destX := destX max: leftMargin].
	kern := kern - font baseKern.

	"Install various parameters from the font."
	spaceWidth := font widthOf: Space.
	xTable := font xTable.
	stopConditions := DefaultStopConditions.! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
setFont: fontNumber
	"Set the font by number from the textStyle."

	self setActualFont: (textStyle fontAt: fontNumber)! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
text: t textStyle: ts 
	text := t.
	textStyle := ts! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
textColor: ignored
	"Overridden in DisplayScanner"! !

!CharacterScanner methodsFor: 'initialize' stamp: 'lr 7/4/2009 10:42'!
wantsColumnBreaks: aBoolean 
	wantsColumnBreaks := aBoolean! !

CharacterScanner subclass: #CompositionScanner
	instanceVariableNames: 'spaceX spaceIndex lineHeight baseline lineHeightAtSpace baselineAtSpace'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CompositionScanner commentStamp: '<historical>' prior: 0!
CompositionScanners are used to measure text and determine where line breaks and space padding should occur.!

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:13'!
columnBreak

	"Answer true. Set up values for the text line interval currently being 
	composed."

	pendingKernX := 0.
	line stop: lastIndex.
	spaceX := destX.
	line paddingWidth: rightMargin - spaceX.
	^true! !

!CompositionScanner methodsFor: 'scanning' stamp: 'nice 3/16/2010 20:42'!
composeFrom: startIndex inRectangle: lineRectangle firstLine: firstLine leftSide: leftSide rightSide: rightSide 
	"Answer an instance of TextLineInterval that represents the next line in the paragraph."
	"Set up margins"
	| runLength stopCondition |
	leftMargin := lineRectangle left.
	leftSide ifTrue: 
		[ leftMargin := leftMargin + (firstLine 
				ifTrue: [ textStyle firstIndent ]
				ifFalse: [ textStyle restIndent ]) ].
	destX := spaceX := leftMargin.
	rightMargin := lineRectangle right.
	rightSide ifTrue: [ rightMargin := rightMargin - textStyle rightIndent ].
	lastIndex := startIndex.	"scanning sets last index"
	destY := lineRectangle top.
	lineHeight := baseline := 0.	"Will be increased by setFont"
	self setStopConditions.	"also sets font"
	runLength := text runLengthFor: startIndex.
	runStopIndex := (lastIndex := startIndex) + (runLength - 1).
	line := (TextLine 
		start: lastIndex
		stop: 0
		internalSpaces: 0
		paddingWidth: 0) rectangle: lineRectangle.
	spaceCount := 0.
	self handleIndentation.
	leftMargin := destX.
	line leftMargin: leftMargin.
	[
		stopCondition := self 
			scanCharactersFrom: lastIndex
			to: runStopIndex
			in: text string
			rightX: rightMargin
			stopConditions: stopConditions
			kern: kern.
		"See setStopConditions for stopping conditions for composing."
		self perform: stopCondition
	] whileFalse.
	^ line
		lineHeight: lineHeight + textStyle leading
		baseline: baseline + textStyle leading! !

!CompositionScanner methodsFor: 'scanning' stamp: 'nice 3/16/2010 20:43'!
composeLine: lineIndex fromCharacterIndex: startIndex inParagraph: aParagraph 
	"Answer an instance of TextLineInterval that represents the next line in the paragraph."
	| runLength stopCondition |
	destX := spaceX := leftMargin := aParagraph leftMarginForCompositionForLine: lineIndex.
	destY := 0.
	rightMargin := aParagraph rightMarginForComposition.
	leftMargin >= rightMargin ifTrue: [ self error: 'No room between margins to compose' ].
	lastIndex := startIndex.	"scanning sets last index"
	lineHeight := textStyle lineGrid.	"may be increased by setFont:..."
	baseline := textStyle baseline.
	self setStopConditions.	"also sets font"
	self handleIndentation.
	runLength := text runLengthFor: startIndex.
	runStopIndex := (lastIndex := startIndex) + (runLength - 1).
	line := TextLineInterval 
		start: lastIndex
		stop: 0
		internalSpaces: 0
		paddingWidth: 0.
	spaceCount := 0.
	[
		stopCondition := self 
			scanCharactersFrom: lastIndex
			to: runStopIndex
			in: text string
			rightX: rightMargin
			stopConditions: stopConditions
			kern: kern.
		"See setStopConditions for stopping conditions for composing."
		self perform: stopCondition
	] whileFalse.
	^ line
		lineHeight: lineHeight + textStyle leading
		baseline: baseline + textStyle leading! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'nice 11/16/2009 09:23'!
cr
	"Answer true. Set up values for the text line interval currently being 
	composed."

	pendingKernX := 0.
	(lastIndex < text size and: [(text at: lastIndex) = CR and: [(text at: lastIndex+1) = Character lf]]) ifTrue: [lastIndex := lastIndex + 1].
	line stop: lastIndex.
	spaceX := destX.
	line paddingWidth: rightMargin - spaceX.
	^true! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:14'!
crossedX
	"There is a word that has fallen across the right edge of the composition 
	rectangle. This signals the need for wrapping which is done to the last 
	space that was encountered, as recorded by the space stop condition."

	pendingKernX := 0.
	spaceCount >= 1 ifTrue:
		["The common case. First back off to the space at which we wrap."
		line stop: spaceIndex.
		lineHeight := lineHeightAtSpace.
		baseline := baselineAtSpace.
		spaceCount := spaceCount - 1.
		spaceIndex := spaceIndex - 1.

		"Check to see if any spaces preceding the one at which we wrap.
			Double space after punctuation, most likely."
		[(spaceCount > 1 and: [(text at: spaceIndex) = Space])]
			whileTrue:
				[spaceCount := spaceCount - 1.
				"Account for backing over a run which might
					change width of space."
				font := text fontAt: spaceIndex withStyle: textStyle.
				spaceIndex := spaceIndex - 1.
				spaceX := spaceX - (font widthOf: Space)].
		line paddingWidth: rightMargin - spaceX.
		line internalSpaces: spaceCount]
	ifFalse:
		["Neither internal nor trailing spaces -- almost never happens."
		lastIndex := lastIndex - 1.
		[destX <= rightMargin]
			whileFalse:
				[destX := destX - (font widthOf: (text at: lastIndex)).
				lastIndex := lastIndex - 1].
		spaceX := destX.
		line paddingWidth: rightMargin - destX.
		line stop: (lastIndex max: line first)].
	^true! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
endOfRun
	"Answer true if scanning has reached the end of the paragraph. 
	Otherwise step conditions (mostly install potential new font) and answer 
	false."
	| runLength |
	lastIndex = text size 
		ifTrue: 
			[ line stop: lastIndex.
			spaceX := destX.
			line paddingWidth: rightMargin - destX.
			^ true ]
		ifFalse: 
			[ runLength := text runLengthFor: (lastIndex := lastIndex + 1).
			runStopIndex := lastIndex + (runLength - 1).
			self setStopConditions.
			^ false ]! !

!CompositionScanner methodsFor: 'intialize-release' stamp: 'ar 5/17/2000 19:14'!
forParagraph: aParagraph
	"Initialize the receiver for scanning the given paragraph."

	self
		initializeFromParagraph: aParagraph
		clippedBy: aParagraph clippingRectangle.
! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	"Workaround: The following should really use #textAnchorType"
	| descent |
	anchoredMorph relativeTextAnchorPosition ifNotNil: [ ^ true ].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: 
		[ "It doesn't fit"
		"But if it's the first character then leave it here"
		lastIndex < line first ifFalse: 
			[ line stop: lastIndex - 1.
			^ false ] ].
	descent := lineHeight - baseline.
	lineHeight := lineHeight max: anchoredMorph height.
	baseline := lineHeight - descent.
	line stop: lastIndex.
	^ true! !

!CompositionScanner methodsFor: 'accessing' stamp: 'ar 1/8/2000 14:35'!
rightX
	"Meaningful only when a line has just been composed -- refers to the 
	line most recently composed. This is a subtrefuge to allow for easy 
	resizing of a composition rectangle to the width of the maximum line. 
	Useful only when there is only one line in the form or when each line 
	is terminated by a carriage return. Handy for sizing menus and lists."

	^spaceX! !

!CompositionScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
setActualFont: aFont 
	"Keep track of max height and ascent for auto lineheight"
	| descent |
	super setActualFont: aFont.
	lineHeight == nil 
		ifTrue: 
			[ descent := font descent.
			baseline := font ascent.
			lineHeight := baseline + descent ]
		ifFalse: 
			[ descent := lineHeight - baseline max: font descent.
			baseline := baseline max: font ascent.
			lineHeight := lineHeight max: baseline + descent ]! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
setFont
	super setFont.
	stopConditions == DefaultStopConditions ifTrue: [ stopConditions := stopConditions copy ].
	stopConditions 
		at: Space asciiValue + 1
		put: #space.
	wantsColumnBreaks == true ifTrue: 
		[ stopConditions 
			at: TextComposer characterForColumnBreak asciiValue + 1
			put: #columnBreak ]! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/8/2000 14:37'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:14'!
space
	"Record left x and character index of the space character just encounted. 
	Used for wrap-around. Answer whether the character has crossed the 
	right edge of the composition rectangle of the paragraph."

	pendingKernX := 0.
	spaceX := destX.
	destX := spaceX + spaceWidth.
	spaceIndex := lastIndex.
	lineHeightAtSpace := lineHeight.
	baselineAtSpace := baseline.
	lastIndex := lastIndex + 1.
	spaceCount := spaceCount + 1.
	destX > rightMargin ifTrue: 	[^self crossedX].
	^false
! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:15'!
tab
	"Advance destination x according to tab settings in the paragraph's 
	textStyle. Answer whether the character has crossed the right edge of 
	the composition rectangle of the paragraph."

	pendingKernX := 0.
	destX := textStyle
				nextTabXFrom: destX leftMargin: leftMargin rightMargin: rightMargin.
	destX > rightMargin ifTrue:	[^self crossedX].
	lastIndex := lastIndex + 1.
	^false
! !

CompositionScanner subclass: #SegmentScanner
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!SegmentScanner methodsFor: 'as yet unclassified' stamp: 'lr 7/4/2009 10:42'!
setFont
	super setFont.
	"Make a local copy of stop conditions so we don't modify the default"
	stopConditions == DefaultStopConditions ifTrue: [ stopConditions := stopConditions copy ].
	stopConditions 
		at: Space asciiValue + 1
		put: nil! !

CharacterScanner subclass: #DisplayScanner
	instanceVariableNames: 'bitBlt lineY runX foregroundColor backgroundColor fillBlt lineHeight paragraph paragraphColor morphicOffset ignoreColorChanges'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!DisplayScanner commentStamp: '<historical>' prior: 0!
My instances are used to scan text and display it on the screen or in a hidden form.!

!DisplayScanner class methodsFor: 'queries' stamp: 'ar 5/17/2000 17:39'!
defaultFont
	^ TextStyle defaultFont! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'nice 11/16/2009 15:02'!
cr
	"When a carriage return is encountered, simply increment the pointer 
	into the paragraph."

	pendingKernX := 0.
	(lastIndex < text size and: [(text at: lastIndex) = CR and: [(text at: lastIndex+1) = Character lf]])
		ifTrue: [lastIndex := lastIndex + 2]
		ifFalse: [lastIndex := lastIndex + 1].
	^false! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'di 9/3/2000 16:24'!
crossedX
	"This condition will sometimes be reached 'legally' during display, when, 
	for instance the space that caused the line to wrap actually extends over 
	the right boundary. This character is allowed to display, even though it 
	is technically outside or straddling the clipping ectangle since it is in 
	the normal case not visible and is in any case appropriately clipped by 
	the scanner."

	^ true ! !

!DisplayScanner methodsFor: 'scanning' stamp: 'nice 3/16/2010 19:34'!
displayLine: textLine offset: offset leftInRun: leftInRun 
	"The call on the primitive (scanCharactersFrom:to:in:rightX:) will be interrupted according to an array of stop conditions passed to the scanner at which time the code to handle the stop condition is run and the call on the primitive continued until a stop condition returns true (which means the line has terminated).  leftInRun is the # of characters left to scan in the current run; when 0, it is time to call setStopConditions."
	| stopCondition nowLeftInRun startIndex string lastPos |
	line := textLine.
	morphicOffset := offset.
	lineY := line top + offset y.
	lineHeight := line lineHeight.
	rightMargin := line rightMargin + offset x.
	lastIndex := line first.
	leftInRun <= 0 ifTrue: [ self setStopConditions ].
	leftMargin := (line leftMarginForAlignment: alignment) + offset x.
	destX := runX := leftMargin.
	fillBlt == nil ifFalse: 
		[ "Not right"
		fillBlt
			destX: line left
				destY: lineY
				width: line width left
				height: lineHeight;
			copyBits ].
	lastIndex := line first.
	leftInRun <= 0 
		ifTrue: [ nowLeftInRun := text runLengthFor: lastIndex ]
		ifFalse: [ nowLeftInRun := leftInRun ].
	destY := lineY + line baseline - font ascent.
	runStopIndex := lastIndex + (nowLeftInRun - 1) min: line last.
	spaceCount := 0.
	string := text string.
	[
		startIndex := lastIndex.
		lastPos := destX @ destY.
		stopCondition := self 
			scanCharactersFrom: lastIndex
			to: runStopIndex
			in: string
			rightX: rightMargin
			stopConditions: stopConditions
			kern: kern.
		lastIndex >= startIndex ifTrue: 
			[ font 
				displayString: string
				on: bitBlt
				from: startIndex
				to: lastIndex
				at: lastPos
				kern: kern ].
		"see setStopConditions for stopping conditions for displaying."
		(self perform: stopCondition)
			or: [lastIndex > runStopIndex]
	] whileFalse.
	^ runStopIndex - lastIndex	"Number of characters remaining in the current run"! !

!DisplayScanner methodsFor: 'mvc-compatibility' stamp: 'nice 3/16/2010 19:37'!
displayLines: linesInterval in: aParagraph clippedBy: visibleRectangle 
	"The central display routine. The call on the primitive 
	(scanCharactersFrom:to:in:rightX:) will be interrupted according to an 
	array of stop conditions passed to the scanner at which time the code to 
	handle the stop condition is run and the call on the primitive continued 
	until a stop condition returns true (which means the line has 
	terminated)."
	"leftInRun is the # of characters left to scan in the current run;
		when 0, it is time to call 'self setStopConditions'"
	| leftInRun |
	morphicOffset := 0 @ 0.
	leftInRun := 0.
	self 
		initializeFromParagraph: aParagraph
		clippedBy: visibleRectangle.
	ignoreColorChanges := false.
	paragraph := aParagraph.
	foregroundColor := paragraphColor := aParagraph foregroundColor.
	backgroundColor := aParagraph backgroundColor.
	aParagraph backgroundColor isTransparent 
		ifTrue: [ fillBlt := nil ]
		ifFalse: 
			[ fillBlt := bitBlt copy.	"Blt to fill spaces, tabs, margins"
			fillBlt
				sourceForm: nil;
				sourceOrigin: 0 @ 0.
			fillBlt fillColor: aParagraph backgroundColor ].
	rightMargin := aParagraph rightMarginForDisplay.
	lineY := aParagraph topAtLineIndex: linesInterval first.
	bitBlt destForm 
		deferUpdatesIn: visibleRectangle
		while: 
			[ linesInterval do: 
				[ :lineIndex |
				| string stopCondition startIndex runLength lastPos | 
				line := aParagraph lines at: lineIndex.
				lastIndex := line first.
				self setStopConditions.	" causes an assignment to inst var.  alignment "
				leftMargin := aParagraph 
					leftMarginForDisplayForLine: lineIndex
					alignment: (alignment ifNil: [ textStyle alignment ]).
				destX := runX := leftMargin.
				line := aParagraph lines at: lineIndex.
				lineHeight := line lineHeight.
				fillBlt == nil ifFalse: 
					[ fillBlt
						destX: visibleRectangle left
							destY: lineY
							width: visibleRectangle width
							height: lineHeight;
						copyBits ].
				lastIndex := line first.
				leftInRun <= 0 ifTrue: 
					[ self setStopConditions.	"also sets the font"
					leftInRun := text runLengthFor: line first ].
				destY := lineY + line baseline - font ascent.	"Should have happened in setFont"
				runLength := leftInRun.
				runStopIndex := lastIndex + (runLength - 1) min: line last.
				leftInRun := leftInRun - (runStopIndex - lastIndex + 1).
				spaceCount := 0.
				string := text string.
				self handleIndentation.
				[
					startIndex := lastIndex.
					lastPos := destX @ destY.
					stopCondition := self 
						scanCharactersFrom: lastIndex
						to: runStopIndex
						in: string
						rightX: rightMargin
						stopConditions: stopConditions
						kern: kern.
					lastIndex >= startIndex ifTrue: 
						[ font 
							displayString: string
							on: bitBlt
							from: startIndex
							to: lastIndex
							at: lastPos
							kern: kern ].
					"see setStopConditions for stopping conditions for displaying."
					self perform: stopCondition
				] whileFalse.
				fillBlt == nil ifFalse: 
					[ fillBlt
						destX: destX
							destY: lineY
							width: visibleRectangle right - destX
							height: lineHeight;
						copyBits ].
				lineY := lineY + lineHeight ] ]! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
endOfRun
	"The end of a run in the display case either means that there is actually 
	a change in the style (run code) to be associated with the string or the 
	end of this line has been reached."
	| runLength |
	lastIndex = line last ifTrue: [ ^ true ].
	runX := destX.
	runLength := text runLengthFor: (lastIndex := lastIndex + 1).
	runStopIndex := lastIndex + (runLength - 1) min: line last.
	self setStopConditions.
	^ false! !

!DisplayScanner methodsFor: 'mvc-compatibility' stamp: 'pavel.krivanek 11/21/2008 16:52'!
initializeFromParagraph: aParagraph clippedBy: clippingRectangle 
	super 
		initializeFromParagraph: aParagraph
		clippedBy: clippingRectangle.
	bitBlt := UIManager default grafPort toForm: aParagraph destinationForm.
	bitBlt
		sourceX: 0;
		width: 0.	"Init BitBlt so that the first call to a primitive will not fail"
	bitBlt combinationRule: (Display depth = 1 
			ifTrue: [ aParagraph rule ]
			ifFalse: [ Form paint ]).
	bitBlt colorMap: (Bitmap 
			with: 0
			with: (bitBlt destForm pixelValueFor: aParagraph foregroundColor)).	"Assumes 1-bit deep fonts"
	bitBlt clipRect: clippingRectangle! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 12:51'!
paddedSpace
	"Each space is a stop condition when the alignment is right justified. 
	Padding must be added to the base width of the space according to 
	which space in the line this space is and according to the amount of 
	space that remained at the end of the line when it was composed."

	spaceCount := spaceCount + 1.
	destX := destX + spaceWidth + (line justifiedPadFor: spaceCount font: font).
	lastIndex := lastIndex + 1.
	^ false! !

!DisplayScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	anchoredMorph relativeTextAnchorPosition ifNotNil: 
		[ anchoredMorph position: anchoredMorph relativeTextAnchorPosition + (anchoredMorph owner textBounds origin x @ 0) - (0 @ morphicOffset y) + (0 @ lineY).
		^ true ].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: [ ^ false ].
	anchoredMorph isMorph 
		ifTrue: 
			[ anchoredMorph position: (destX - anchoredMorph width) @ lineY - morphicOffset ]
		ifFalse: 
			[ destY := lineY.
			runX := destX.
			anchoredMorph 
				displayOn: bitBlt destForm
				at: (destX - anchoredMorph width) @ destY
				clippingBox: bitBlt clipRect ].
	^ true! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
plainTab
	| oldX |
	oldX := destX.
	super plainTab.
	fillBlt == nil ifFalse: 
		[ fillBlt
			destX: oldX
				destY: destY
				width: destX - oldX
				height: font height;
			copyBits ]! !

!DisplayScanner methodsFor: 'private' stamp: 'ar 5/17/2000 19:26'!
setDestForm: df
	bitBlt setDestForm: df.! !

!DisplayScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setFont
	foregroundColor := paragraphColor.
	super setFont.	"Sets font and emphasis bits, and maybe foregroundColor"
	font 
		installOn: bitBlt
		foregroundColor: foregroundColor
		backgroundColor: Color transparent.
	text ifNotNil: [ destY := lineY + line baseline - font ascent ]! !

!DisplayScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setPort: aBitBlt 
	"Install the BitBlt to use"
	bitBlt := aBitBlt.
	bitBlt
		sourceX: 0;
		width: 0.	"Init BitBlt so that the first call to a primitive will not fail"
	bitBlt sourceForm: nil	"Make sure font installation won't be confused"! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'yo 10/4/2002 20:43'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont.
	self setConditionArray: (alignment = Justified ifTrue: [#paddedSpace]).

"
	alignment = Justified ifTrue: [
		stopConditions == DefaultStopConditions 
			ifTrue:[stopConditions := stopConditions copy].
		stopConditions at: Space asciiValue + 1 put: #paddedSpace]
"! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
tab
	self plainTab.
	lastIndex := lastIndex + 1.
	^ false! !

!DisplayScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
text: t textStyle: ts foreground: foreColor background: backColor fillBlt: blt ignoreColorChanges: shadowMode 
	text := t.
	textStyle := ts.
	foregroundColor := paragraphColor := foreColor.
	(backgroundColor := backColor) isTransparent ifFalse: 
		[ fillBlt := blt.
		fillBlt fillColor: backgroundColor ].
	ignoreColorChanges := shadowMode! !

!DisplayScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
textColor: textColor 
	ignoreColorChanges ifTrue: [ ^ self ].
	foregroundColor := textColor! !

Object subclass: #StandardFonts
	instanceVariableNames: ''
	classVariableNames: 'ButtonFont CodeFont HaloFont ListFont MenuFont WindowTitleFont'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 09:36'!
balloonFont
	^BalloonMorph balloonFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 09:36'!
balloonFont: aFont
	BalloonMorph balloonFont: aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/28/2009 21:47'!
buttonFont
	^ ButtonFont ifNil: [ButtonFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:07'!
buttonFont: aFont
	 ButtonFont := aFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:50'!
chooseBalloonFont
	self chooseStandardFont: #balloonFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:51'!
chooseButtonFont
	self chooseStandardFont: #buttonFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:51'!
chooseCodeFont
	self chooseStandardFont: #codeFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:48'!
chooseDefaultFont
	self chooseStandardFont: #defaultFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/18/2009 13:58'!
chooseFontWithTitle: aTitle for: aReceiver setSelector: aSetSelector getSelector: aGetSelector
	FontChooser
		openWithWindowTitle: aTitle
		for: aReceiver
		setSelector: aSetSelector
		getSelector: aGetSelector! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:49'!
chooseHaloFont
	self chooseStandardFont: #haloFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:49'!
chooseListFont
	self chooseStandardFont: #listFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:50'!
chooseMenuFont
	self chooseStandardFont: #menuFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/18/2009 13:59'!
chooseStandardFont: aSelector 
	self
		chooseFontWithTitle: (self systemFontLabelOf: aSelector)	, '...'
		for: self
		setSelector: (aSelector , ':') asSymbol
		getSelector: aSelector! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:50'!
chooseWindowTitleFont
	self chooseStandardFont: #windowTitleFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:22'!
codeFont
	^ CodeFont ifNil: [CodeFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:22'!
codeFont: aFont
	 CodeFont := aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 19:15'!
defaultFont
	^ TextStyle defaultFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'StephaneDucasse 11/26/2010 19:31'!
defaultFont: aFont
	"Establish the default text font and style"

	| aStyle newDefaultStyle |
	aFont ifNil: [^ StandardFonts setFontsFromSpec: (Array with: {#defaultFont:. StrikeFont accunyKey. 12})].
	aStyle := aFont textStyle ifNil: [^ self].
	newDefaultStyle := aStyle copy.
	newDefaultStyle defaultFontIndex: (aStyle fontIndexOf: aFont).
	TextSharedInformation at: #DefaultTextStyle put: newDefaultStyle.
! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/9/2009 09:24'!
getCurrentFontsSpec
	"self getCurrentFontsSpec"
	^ self systemFontAspects collect: [:aspect | | font |
		font := self perform: aspect.
		Array with: (aspect, ':') asSymbol with: font familyName with: font pointSize]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:19'!
haloFont
	^ HaloFont ifNil: [HaloFont := StandardFonts defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:20'!
haloFont: aFont
	HaloFont := aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:24'!
listFont
	^ ListFont ifNil: [ListFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:24'!
listFont: aFont
	 ListFont := aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 11:55'!
menuFont
	^ MenuFont ifNil: [MenuFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'StephaneDucasse 7/2/2010 19:01'!
menuFont: aFont
	aFont ~= MenuFont 
		ifTrue: [ MenuFont := aFont ].! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/19/2009 17:56'!
printStandardSystemFonts
	"self printStandardSystemFonts"

	| string |
	string := String streamContents: [ :s |

	StandardFonts systemFontAspects do: [:selector |
		| font |
		font := StandardFonts perform: selector.
		s
			nextPutAll: selector; space;
			nextPutAll: font familyName; space;
			nextPutAll: (AbstractFont emphasisStringFor: font emphasis);
			nextPutAll: ' points: ';
			print: font pointSize;
			nextPutAll: ' height: ';
			print: font height;
			cr
		]].

	(StringHolder new)
		contents: string;
		openLabel: 'Current system font settings' translated.
! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 18:06'!
restoreDefaultFonts
	"Since this is called from menus, we can take the opportunity to prompt for missing font styles."
	"
	self restoreDefaultFonts
	"	
	self setFontsFromSpec: 
		#( #(#defaultFont: 'Bitmap DejaVu Sans' 9 ) 
			#(#codeFont: 'Bitmap DejaVu Sans' 9 ) 
			#(#listFont: 'Bitmap DejaVu Sans' 9 ) 
			#(#menuFont: 'Bitmap DejaVu Sans' 9 ) 
			#(#windowTitleFont: 'Bitmap DejaVu Sans' 12 ) 
			#(#balloonFont: 'Accuny' 10 ) 
			#(#haloFont: 'Accuny' 10 ) 
			#(#buttonFont: 'Bitmap DejaVu Sans' 9 ) ).
! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/9/2009 08:46'!
setAllStandardFontsTo: aFont
	self systemFontAspects do: [:selector | StandardFonts perform: (selector, ':') asSymbol with: aFont].
	 
! !

!StandardFonts class methodsFor: 'menu' stamp: 'MarcusDenker 8/15/2010 13:11'!
setDefaultFonts: defaultFontsSpec
	"Since this is called from menus, we can take the opportunity to prompt for missing font styles."

	| fontNames map emphases |
	fontNames := defaultFontsSpec collect: [:array | array second].
	map := IdentityDictionary new.
	emphases := IdentityDictionary new.
	fontNames do: [:originalName | | decoded style response |
		decoded := TextStyle decodeStyleName: originalName.
		style := map at: originalName put: (TextStyle named: decoded second).
		emphases at: originalName put: decoded first.
		style ifNil: [map at: originalName put: (TextStyle default) ]].
	defaultFontsSpec do: [:triplet | self
		perform: triplet first
		with: (((map at: triplet second) fontOfPointSize: triplet third) emphasized: (emphases at: triplet second))]! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:52'!
setDemoFonts
	"StandardFonts setDemoFonts"
	|size font codeFont titleFont|

	size := UIManager default request: 'Base font size?' initialAnswer: '14'.
	size isEmptyOrNil ifTrue: [^ self].
	size := size asInteger.
	(size isNil or: [size <= 0]) ifTrue: [^ self].
	font := LogicalFont familyName: 'DejaVu Sans' pointSize: size.
	codeFont := LogicalFont familyName: 'DejaVu Sans Mono' pointSize: size.
	titleFont := LogicalFont familyName: 'DejaVu Serif' pointSize: size.

	StandardFonts
		listFont: font;
		menuFont: font;
		codeFont: codeFont;
		buttonFont: font;
		defaultFont: font;
		windowTitleFont: titleFont.! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/9/2009 09:22'!
setFontsFromSpec: fontsSpec
	"set some fonts by interpreting fontsSpec.
	fontsSpec is a collection.
	Each fontsSpec element is an array with 3 elements: 
	(1) the font accessor (i.e. #codeFont:)
	(2) the font family (i.e. 'Arial')
	(3) the font point size.
	To see an example of fontsSpec, inspect the result of 
	StandardFonts getCurrentFontsSpec.
	"
	| fontNames map emphases |
	fontNames := fontsSpec
				collect: [:array | array second].
	map := IdentityDictionary new.
	emphases := IdentityDictionary new.
	fontNames
		do: [:originalName | 
			| decoded style response |
			decoded := TextStyle decodeStyleName: originalName.
			style := map
						at: originalName
						put: (TextStyle named: decoded second).
			emphases at: originalName put: decoded first.
			style
				ifNil: [map at: originalName put: TextStyle default]].
	fontsSpec
		do: [:triplet | self
				perform: triplet first
				with: (((map at: triplet second)
						fontOfPointSize: triplet third)
						emphasized: (emphases at: triplet second))]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/18/2009 13:49'!
systemFontAspects
	^ self systemFontAspectsAndLabels collect: [:assoc | assoc key]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/18/2009 13:53'!
systemFontAspectsAndLabels
	^ {#defaultFont -> 'Default font' translated. #listFont -> 'List font'  translated. #menuFont -> 'Menu font'  translated. #windowTitleFont -> 'Window title'  translated. #balloonFont -> 'Balloon help font'  translated. #codeFont -> 'Code font'  translated. #buttonFont -> 'Button font'  translated. #haloFont -> 'Halo font'  translated } ! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/18/2009 13:51'!
systemFontLabelOf: aSelector
	^ (self systemFontAspectsAndLabels detect: [:assoc | assoc key = aSelector] ifNone: [^ '']) value! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/19/2009 18:10'!
systemFontLabels
	^ self systemFontAspectsAndLabels collect: [:assoc | assoc value]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:21'!
windowTitleFont
	^ WindowTitleFont ifNil: [WindowTitleFont := StandardFonts defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:21'!
windowTitleFont: aFont
	WindowTitleFont := aFont! !

Object subclass: #TextPrinter
	instanceVariableNames: 'form para paperSize landscape resolution depth offset columns docTitle noHeader noFooter'
	classVariableNames: 'DefaultPaperSize DefaultTextPrinter'
	poolDictionaries: ''
	category: 'Graphics-Text'!

!TextPrinter class methodsFor: 'accessing' stamp: 'ar 4/30/98 18:31'!
defaultPaperSize
	^DefaultPaperSize! !

!TextPrinter class methodsFor: 'accessing' stamp: 'ar 4/30/98 18:31'!
defaultPaperSize: aPoint
	DefaultPaperSize := aPoint! !

!TextPrinter class methodsFor: 'accessing' stamp: 'lr 7/4/2009 10:42'!
defaultTextPrinter
	"This is the global default TextPrinter instance."
	DefaultTextPrinter isNil ifTrue: [ DefaultTextPrinter := self new ].
	^ DefaultTextPrinter! !

!TextPrinter class methodsFor: 'initialization' stamp: 'ar 4/30/98 18:30'!
initialize
	"TextPrinter initialize"
	self defaultPaperSize: self paperSizeA4.! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:30'!
mm2in: aPoint
	"Convert aPoint from millimeters to inches"
	^aPoint / 25.4! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSize10x14
	^10.0@14.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSize11x17
	^11.0@17.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeA3
	^self mm2in: 297@420! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeA4
	^self mm2in: 210@297! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeA5
	^self mm2in: 148@210! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeB4
	^self mm2in: 250@354! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeB5
	^self mm2in: 182@257! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeCSheet
	^17.0@22.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeDSheet
	^22.0@34.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeESheet
	^34.0@44.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope10
	^4.125@9.5
! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope11
	^4.5@10.375! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope12
	^4.75@11! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope14
	^5.0@11.5! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope9
	^3.875@8.875! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeB4
	^self mm2in: 250@353! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeB5
	^self mm2in: 176@250! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeB6
	^self mm2in: 176@125! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeC3
	^self mm2in: 324@458! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC4
	^self mm2in: 229@324! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC5
	^self mm2in: 162@229! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC6
	^self mm2in: 114@162! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC65
	^self mm2in: 114@229! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFanfoldGerman
	"German standard fanfold"
	^8.5@12.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFanfoldLegalGerman
	"German legal fanfold"
	^8.5@13.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFanfoldUS
	"US standard fanfold"
	^14.875@11.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFolio
	^8.5@13.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeLegal
	^8.5@14.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeLetter
	^8.5@11.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeNote
	^8.5@11.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:30'!
paperSizeTabloid
	^11.0@17.0! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
bestColor
	"Set the reproduction quality to true color"
	depth := 32.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
blackAndWhite
	"Set the reproduction quality to black and white"
	depth := 1.! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:25'!
columnRect: n
	"Return a rectangle describing the n-th column"
	| area left right |
	area := self textArea.
	left := area left + ((n-1) * self columnWidth).
	left := left + ((n-1) * self columnSkip).
	right := left + self columnWidth.
	^(self in2pix: left @ area top) corner: 
		(self in2pix: right @ area bottom)! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:20'!
columnSkip
	"Return the separating space between two columns in inches"
	^0.2! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:21'!
columnWidth
	^(self textWidth - ((self columns-1) * self columnSkip)) / self columns! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:50'!
columns
	^columns! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:50'!
columns: aNumber
	columns := aNumber asInteger max: 1.! !

!TextPrinter methodsFor: 'initialize' stamp: 'ar 4/30/98 19:26'!
defaultPaperSize
	"Return the default paper size (inches) for printing"
	^self class defaultPaperSize! !

!TextPrinter methodsFor: 'initialize' stamp: 'nk 4/2/2004 11:32'!
defaultResolution
	"Return the default resolution (DPI) for printing"
	^TextStyle pixelsPerInch asPoint! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 20:14'!
documentTitle
	^docTitle! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 20:14'!
documentTitle: aString
	docTitle := aString! !

!TextPrinter methodsFor: 'printing' stamp: 'ar 4/30/98 20:41'!
flushPage
	"The current page has been set up. Send it to the printer."
	form primPrintHScale: self resolution x vScale: self resolution y landscape: self landscape.
	"Uncomment the following for testing"
	"form displayOn: Display. (Delay forSeconds: 5) wait."
! !

!TextPrinter methodsFor: 'footer' stamp: 'ar 4/30/98 19:23'!
footerHeight
	"Return the (additional) height of the footer in inches."
	self noFooter ifTrue:[^0.0].
	^(self pix2in: 0@TextStyle default lineGrid) y * 2! !

!TextPrinter methodsFor: 'footer' stamp: 'ar 4/30/98 20:11'!
footerParagraph
	"Return a paragraph for the footer"
	| fPara rect |
	fPara := Paragraph new.
	fPara destinationForm: form.
	rect := (self in2pix: self textArea bottomLeft) corner: 
				(self in2pix: self textArea bottomRight + (0.0@self footerHeight)).
	fPara clippingRectangle: rect.
	fPara compositionRectangle: rect.
	^fPara! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:29'!
formatColumn: columnNum startingWith: anIndex
	"Format a new column starting at the given string index. Return the string index indicating the start of the next column or nil if no more columns need printing."
	| colRect blk |
	colRect := self columnRect: columnNum.
	anIndex > 1 ifTrue:[para text: (para text copyFrom: anIndex to: para text size)].
	para compositionRectangle: colRect.
	para clippingRectangle: colRect.
	para composeAll.
	para displayOn: form.
	para visibleRectangle corner y <= colRect extent y ifTrue:[^nil].
	"More columns -- find the character block of the last line and adjust clip rect"
	blk := para characterBlockAtPoint: para visibleRectangle bottomLeft.
	para clearVisibleRectangle. "Make sure that the background is clean"
	para clippingRectangle: (colRect topLeft corner: colRect right@blk top).
	para displayOn: form.
	^blk stringIndex.! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:29'!
formatPage: pageNum startingWith: anIndex
	"Format a new page starting at the given string index. Return the string index indicating the start of the next page or nil if no more pages need printing."
	| nextIndex |
	nextIndex := anIndex.
	1 to: self columns do:[:i|
		nextIndex := self formatColumn: i startingWith: nextIndex.
		nextIndex isNil ifTrue:[^nil].
	].
	^nextIndex! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
goodColor
	"Set the reproduction quality to 8 bit color depth"
	depth := 8.! !

!TextPrinter methodsFor: 'header' stamp: 'ar 4/30/98 19:23'!
headerHeight
	"Return the (additional) height of the header in inches."
	self noHeader ifTrue:[^0.0].
	^(self pix2in: 0@TextStyle default lineGrid) y * 2! !

!TextPrinter methodsFor: 'header' stamp: 'ar 4/30/98 20:11'!
headerParagraph
	"Return a paragraph for the footer"
	| hPara rect |
	hPara := Paragraph new.
	hPara destinationForm: form.
	rect := (self in2pix: self textArea topLeft - (0.0@self headerHeight)) corner: 
				(self in2pix: self textArea topRight).
	hPara clippingRectangle: rect.
	hPara compositionRectangle: rect.
	^hPara! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:39'!
in2mm: aPoint
	"Convert aPoint from millimeters to inches"
	^aPoint * 25.4! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:38'!
in2pix: aPoint
	"Convert aPoint from inches to actual pixels"
	^(aPoint * self resolution) rounded! !

!TextPrinter methodsFor: 'initialize' stamp: 'adrian_lienhard 7/18/2009 16:03'!
initialize
	self paperSize: self defaultPaperSize.
	self resolution: self defaultResolution.
	self blackAndWhite.
	self landscape: false.
	self offsetRect: (1.0@1.0 corner: 1.0@1.0).
	self columns: 1.
	self noHeader: false.
	self noFooter: false.
	self documentTitle: 'Pharo Document (from ', Date today printString,')'.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
landscape
	^landscape! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
landscape: aBoolean
	landscape := aBoolean! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:38'!
mm2in: aPoint
	"Convert aPoint from millimeters to inches"
	^aPoint / 25.4! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:40'!
mm2pix: aPoint
	"Convert aPoint from millimeters to actual pixels"
	^self in2pix: (self mm2in: aPoint)! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:23'!
noFooter
	^noFooter! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:22'!
noFooter: aBoolean
	"Turn off footer printing"
	noFooter := aBoolean.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:22'!
noHeader
	^noHeader! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:22'!
noHeader: aBoolean
	"Turn off header printing"
	noHeader := aBoolean.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:27'!
offsetRect
	^offset! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:27'!
offsetRect: aRectangle
	"Set the offset rectangle"
	offset := aRectangle! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
paperSize
	^paperSize! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
paperSize: aPoint
	paperSize := aPoint! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:39'!
pix2in: aPoint
	"Convert aPoint from a pixel value to inches"
	^aPoint / self resolution! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:40'!
pix2mm: aPoint
	"Convert aPoint from a pixel value to millimeters"
	^self in2mm: (self pix2in: aPoint)! !

!TextPrinter methodsFor: 'private' stamp: 'ar 4/30/98 19:40'!
pixelSize
	"Return the size of the page in pixels"
	^self in2pix: (self realPaperSize)! !

!TextPrinter methodsFor: 'footer' stamp: 'ar 4/30/98 19:24'!
printFooter: pageNumber
	"Print the footer for the given page number"
	| fPara |
	self noFooter ifTrue:[^self].
	fPara := self footerParagraph.
	fPara centered.
	fPara text: ('Page ', pageNumber printString) asText.
	fPara displayOn: form.! !

!TextPrinter methodsFor: 'header' stamp: 'ar 4/30/98 19:23'!
printHeader: pageNumber
	"Print the header for the given page number"
	| fPara |
	self noHeader ifTrue:[^self].
	fPara := self headerParagraph.
	fPara centered.
	fPara text: self documentTitle asText.
	fPara displayOn: form.! !

!TextPrinter methodsFor: 'printing' stamp: 'ar 4/30/98 19:19'!
printParagraph
	| pageNum nextIndex |
	para destinationForm: form.
	pageNum := 1.
	nextIndex := 1.
	[form fillColor: Color white.
	self printHeader: pageNum.
	self printFooter: pageNum.
	nextIndex := self formatPage: pageNum startingWith: nextIndex.
	self flushPage.
	nextIndex isNil] whileFalse:[pageNum := pageNum + 1].! !

!TextPrinter methodsFor: 'printing' stamp: 'ar 4/30/98 18:55'!
printText: aText
	"Print aText"
	form isNil ifTrue:[
		form := Form extent: self pixelSize depth: depth.
	].
	para := Paragraph withText: aText asText.
	Cursor wait showWhile:[
		self printParagraph.
	].! !

!TextPrinter methodsFor: 'private' stamp: 'ar 4/30/98 19:39'!
realPaperSize
	^self landscape
		ifTrue:[self paperSize y @ self paperSize x]
		ifFalse:[self paperSize]! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:43'!
resolution
	^resolution! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:43'!
resolution: aPoint
	resolution := aPoint! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:58'!
textArea
	^(self offsetRect origin + (0.0@self headerHeight)) corner:
		(self realPaperSize - self offsetRect corner - (0.0@self footerHeight))! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:23'!
textWidth
	^self textArea extent x! !

Object subclass: #TextStopConditions
	instanceVariableNames: 'stops crossedX endOfRun'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!TextStopConditions commentStamp: 'nice 3/8/2010 14:31' prior: 0!
A TextStopConditions is a private helper class for text composition (See class CharacterScanner and subclasses).
It maps some characters controling the text layout (like carriage return, line feeds, space and tabulation) to some selector representing the action to be performed by the CharacterScanner.
By default, a TextStopConditions does not map any action to control character; instances must be properly initialized by sending #at:put: messages.

For example, inter-word spacing can be adjusted so as to obtain "justified" paragraphs. See implementors of #paddedSpace, #cr, #space, #columnBreak for example of special character actions.

A TextStopConditions also store two selectors for mapping the actions to be taken when:
- end of run is encountered;
- text overflows horizontal composition bounds.
These actions are by default #endOfRun and #crossedX (see implementors of these messages), but can enventually be changed using #endOfRun: and #crossedX:.

In a text (see class Text), the "runs" are used to store text style attributes, so an "end of run" event probably means some action in the textcomposer should be taken to change the font.

TextStopConditions current implementation can only map 256 character codes (from 1 to 256). It is the composer responsibility to encode the character before sending #at:. Presumably, the composer will use the character codePoint + 1 (see implementors of #codePoint).
If this is not sufficient, then this class could be changed to use a Dictionary or a LargeTable

Historically, the EndOfRun and CrossedX were two TextConstant of value 257 and 258, which did occupy corresponding slots in the stops array. Since these are valid character codePoint, this usage has been deprecated. However, because any error in text composition would have catastrophic consequences (unresponsive user interface), backward compatibility with obsolete historical code is still maintained internally which is why the stops array has a sze of 258.

Instance Variables
	crossedX:		<Symbol | nil>
	endOfRun:		<Symbol | nil>
	stops:		<Array of: Symbol | nil>

crossedX
	- selector to perform when the composed text overflows X composition bound

endOfRun
	- selector to perform at end of run

stops
	- an array mapping character code (codePoint + 1) to special actions, or nil if character is to be rendered normally
!

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:01'!
at: anInteger
	"Answer the special action associated with a character of code anInteger, or nil if none.
	The character code should be betxween 1 and 256 (presumably codePoint + 1).
	It can eventually be 257 for endOfRun action, or 258 for crossedX action for backward compatibility with historical squeak versions."
	
	^stops at: anInteger! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:02'!
at: anInteger put: aSymbolOrNil
	"Set the special action associated with a character of code anInteger, or nil if none.
	The character code should be betxween 1 and 256 (presumably codePoint + 1).
	It can eventually be 257 for endOfRun action, or 258 for crossedX action for backward compatibility with historical squeak versions."
	
	anInteger = 257
		ifTrue: [
			self endOfRun: aSymbolOrNil..
			^aSymbolOrNil].
	anInteger = 258
		ifTrue: [
			self crossedX: aSymbolOrNil..
			^aSymbolOrNil].
	^stops at: anInteger put: aSymbolOrNil! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:01'!
crossedX
	"Answer the special action to be performed when crossing composition bounds."
	
	^crossedX! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 12:00'!
crossedX: aSymbolOrNil
	crossedX := aSymbolOrNil.
	
	"Backward compatibility with historical EndOfRun TextConstant handling"
	stops size >= 258 ifTrue: [stops at: 258 put: crossedX]! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:01'!
endOfRun
	"Answer the special action to be performed et end of text."
	
	^endOfRun! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 12:00'!
endOfRun: aSymbolOrNil
	endOfRun := aSymbolOrNil.
	
	"Backward compatibility with historical EndOfRun TextConstant handling"
	stops size >= 257 ifTrue: [stops at: 257 put: endOfRun]! !

!TextStopConditions methodsFor: 'initialize-release' stamp: 'nice 3/8/2010 14:31'!
initialize
	"Initialize the default stop conditions."
	stops := Array new: 258.
	self endOfRun: #endOfRun.
	self crossedX: #crossedX.! !

!TextStopConditions methodsFor: 'copying' stamp: 'nice 4/17/2009 00:52'!
postCopy
	stops := stops copy! !

!TextStopConditions methodsFor: 'initialize-release' stamp: 'nice 2/28/2010 19:26'!
setStops: stopArray
	stops := stopArray! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 23:20'!
size
	"Return the highest code to which an action can be mapped.
	Do not take into account the two slots for endOfRun and crossedX, since they don't map any character code.
	Warning: this message is used for compatibility with obsolete code."
	
	^256! !
TextPrinter initialize!
CharacterScanner initialize!
